#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import contextlib
import warnings
from importlib.util import find_spec

import tensorflow as tf
from neptune.types import File

from neptune_tensorboard.integration.utils import register_pre_hook

IS_GRAPHLIB_AVAILABLE = find_spec("tfgraphviz")
if IS_GRAPHLIB_AVAILABLE:
    import tfgraphviz as tfg

_integrated_with_tensorflow = False

__all__ = ["patch_tensorflow", "NeptuneTensorflowTracker"]


def patch_tensorflow(run, base_namespace):
    global _integrated_with_tensorflow

    if not _integrated_with_tensorflow:
        NeptuneTensorflowTracker(run, base_namespace)
        _integrated_with_tensorflow = True


def track_scalar(name, data, step=None, description=None, run=None, base_namespace=None):
    run[base_namespace]["scalar"][name].append(data)


def track_image(name, data, step=None, run=None, base_namespace=None):
    # expecting 2 or 3 dimensional tensor. If tensor is 4-dimentional,
    # as in https://www.tensorflow.org/api_docs/python/tf/summary/image
    # iterate over first dimension to send all images
    shape = tf.shape(data)
    if len(shape) >= 4:
        for num in range(0, shape[0]):
            run[base_namespace]["image"][name].append(File.as_image(data[num]))
    else:
        run[base_namespace]["image"][name] = File.as_image(data)


def track_text(name, data, step=None, description=None, run=None, base_namespace=None):
    run[base_namespace]["text"][name] = data


def track_graph(graph_data, run=None, base_namespace=None):
    if IS_GRAPHLIB_AVAILABLE:
        graph = tfg.board(graph_data)
        png_bytes = graph.pipe(format="png")
        # There is only one graph
        run[base_namespace]["graph"].upload(File.from_content(png_bytes, extension="png"))
    else:
        # user facing
        warnings.warn("neptune-tensorboard: Skipping model visualization because no tfgraphviz installation was found.")


class NeptuneTensorflowTracker(contextlib.AbstractContextManager):
    def __init__(self, run, base_namespace):
        self.org_scalar = tf.summary.scalar
        self.org_image = tf.summary.image
        self.org_text = tf.summary.text
        self.org_graph = tf.summary.graph

        tf.summary.scalar = register_pre_hook(
            original=tf.summary.scalar, neptune_hook=track_scalar, run=run, base_namespace=base_namespace
        )
        tf.summary.image = register_pre_hook(
            original=tf.summary.image, neptune_hook=track_image, run=run, base_namespace=base_namespace
        )
        tf.summary.text = register_pre_hook(
            original=tf.summary.text, neptune_hook=track_text, run=run, base_namespace=base_namespace
        )
        tf.summary.graph = register_pre_hook(
            original=tf.summary.graph, neptune_hook=track_graph, run=run, base_namespace=base_namespace
        )

    def __exit__(self, exc_type, exc_val, exc_tb):
        tf.summary.scalar = self.org_scalar
        tf.summary.image = self.org_image
        tf.summary.text = self.org_text
        tf.summary.graph = self.org_graph



#
# Copyright (c) 2019, Neptune Labs Sp. z o.o.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

__all__ = ["enable_tensorboard_logging", "__version__"]

import warnings
from contextlib import (
    contextmanager,
    nullcontext,
)
from importlib.util import find_spec

from pkg_resources import parse_version

from neptune_tensorboard.integration.version import __version__

# NOTE: We don't use `importlib.find_spec` here as
#       TF can be installed from multiple packages
#       like tensorflow, tensorflow-macos, etc.
IS_TF_AVAILABLE = True
try:
    import tensorflow as tf  # noqa
except ModuleNotFoundError:
    IS_TF_AVAILABLE = False

IS_PYT_AVAILABLE = find_spec("torch")
IS_TENSORBOARDX_AVAILABLE = find_spec("tensorboardX")

if IS_TF_AVAILABLE:
    MIN_TF_VERSION = "2.0.0-rc0"
    from neptune_tensorboard.integration.tensorflow_integration import (
        NeptuneTensorflowTracker,
        patch_tensorflow,
    )

    def check_tf_version():
        version = "<unknown>"
        try:
            # noinspection PyUnresolvedReferences
            version = parse_version(tf.version.VERSION)

            if version >= parse_version(MIN_TF_VERSION):
                return
        except AttributeError:
            # user facing
            message = (
                f"Unrecognized TensorFlow version: {version}. Please make sure "
                "that the TensorFlow version is >={MIN_TF_VERSION}"
            )
            raise Exception(message)


if IS_PYT_AVAILABLE:
    MIN_PT_VERSION = "1.9.0"
    import torch

    from neptune_tensorboard.integration.pytorch_integration import (
        NeptunePytorchTracker,
        patch_pytorch,
    )

    def check_pytorch_version():
        version = "<unknown>"
        try:
            # noinspection PyUnresolvedReferences
            version = parse_version(torch.__version__)

            if version >= parse_version(MIN_PT_VERSION):
                return
        except AttributeError:
            # user facing
            message = (
                f"Unrecognized PyTorch version: {version}. Please make sure "
                "that the PyTorch version is >={MIN_PT_VERSION}"
            )
            raise Exception(message)


if IS_TENSORBOARDX_AVAILABLE:
    import tensorboardX

    MIN_TBX_VERSION = "2.2.0"
    from neptune_tensorboard.integration.tensorboardx_integration import (
        NeptuneTensorboardXTracker,
        patch_tensorboardx,
    )

    def check_tensorboardx_version():
        version = "<unknown>"
        try:
            # noinspection PyUnresolvedReferences
            version = parse_version(tensorboardX.__version__)

            if version >= parse_version(MIN_TBX_VERSION):
                return
        except AttributeError:
            # user facing
            message = (
                f"Unrecognized tensorboardX version: {version}. Please make sure "
                "that the tensorboardX version is >={MIN_TBX_VERSION}"
            )
            raise Exception(message)


# user facing
FRAMEWORK_NOT_FOUND_WARNING_MSG = (
    "neptune-tensorboard: TensorFlow or PyTorch or tensorboardX was not found, ",
    "please ensure that it is available.",
)


def enable_tensorboard_logging(run, *, base_namespace="tensorboard"):
    if IS_TF_AVAILABLE:
        check_tf_version()
        patch_tensorflow(run, base_namespace)
    if IS_PYT_AVAILABLE:
        check_pytorch_version()
        patch_pytorch(run, base_namespace)
    if IS_TENSORBOARDX_AVAILABLE:
        check_tensorboardx_version()
        patch_tensorboardx(run, base_namespace)

    if not (IS_PYT_AVAILABLE or IS_TF_AVAILABLE or IS_TENSORBOARDX_AVAILABLE):
        warnings.warn(FRAMEWORK_NOT_FOUND_WARNING_MSG)


@contextmanager
def enable_tensorboard_logging_ctx(run, *, base_namespace="tensorboard"):
    tf_tracker, pt_tracker, tbx_tracker = nullcontext(), nullcontext(), nullcontext()
    if IS_TF_AVAILABLE:
        check_tf_version()
        tf_tracker = NeptuneTensorflowTracker(run, base_namespace)

    if IS_PYT_AVAILABLE:
        check_pytorch_version()
        pt_tracker = NeptunePytorchTracker(run, base_namespace)

    if IS_TENSORBOARDX_AVAILABLE:
        check_tensorboardx_version()
        tbx_tracker = NeptuneTensorboardXTracker(run, base_namespace)

    if not (IS_PYT_AVAILABLE or IS_TF_AVAILABLE or IS_TENSORBOARDX_AVAILABLE):
        warnings.warn(FRAMEWORK_NOT_FOUND_WARNING_MSG)

    with pt_tracker, tf_tracker, tbx_tracker:
        yield



import hashlib
import pathlib
import traceback

import click
import neptune
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator

try:
    import tbparse
except ModuleNotFoundError:
    # user facing
    raise ModuleNotFoundError(
        "neptune-tensorboard: tbparse is required for exporting logs. Install it with: pip install tbparse"
    )


def compute_md5_hash(path):
    return hashlib.md5(path.encode()).hexdigest()


class DataSync:
    def __init__(self, project, api_token, path):
        self._project = project
        self._api_token = api_token
        self._path = path

    def run(self):
        # NOTE: Fetching custom_run_ids is not a trivial operation, so
        #       we cache the custom_run_ids here.
        self._existing_custom_run_ids = self._get_existing_neptune_custom_run_ids()
        # Inspect if files correspond to EventFiles.
        for path in pathlib.Path(self._path).glob("**/*tfevents*"):
            try:
                # methods below expect path to be str.
                str_path = str(path)

                # only try export for valid files i.e. files which EventAccumulator
                # can actually read.
                if self._is_valid_tf_event_file(str_path):
                    self._export_to_neptune_run(str_path)
            except Exception as e:
                # user facing
                click.echo("Cannot load run from file '{}'. ".format(path) + "Error: " + str(e))
                try:
                    traceback.print_exc(e)
                except:  # noqa: E722
                    pass

    def _is_valid_tf_event_file(self, path):
        accumulator = EventAccumulator(path)
        accumulator.Reload()
        try:
            accumulator.FirstEventTimestamp()
        except ValueError:
            return False
        return True

    def _get_existing_neptune_custom_run_ids(self):
        with neptune.init_project(project=self._project, api_token=self._api_token) as project:
            try:
                return {run_id for run_id in project.fetch_runs_table().to_pandas()["sys/custom_run_id"].to_list()}
            except KeyError:
                # empty project
                return set()

    def _experiment_exists(self, hash_run_id, run_path):
        return hash_run_id in self._existing_custom_run_ids

    def _export_to_neptune_run(self, path):
        # custom_run_id supports str with max length of 32.
        hash_run_id = compute_md5_hash(path)

        if self._experiment_exists(hash_run_id, self._project):
            # user facing
            click.echo(f"{path} was already synchronized")
            return

        with neptune.init_run(
            custom_run_id=hash_run_id, project=self._project, api_token=self._api_token, capture_hardware_metrics=False
        ) as run:
            run["tensorboard_path"] = path

            namespace_handler = run["tensorboard"]

            # parse events file
            reader = tbparse.SummaryReader(path)

            # Read scalars
            for scalar in reader.scalars.itertuples():
                namespace_handler["scalar"][scalar.tag].append(scalar.value)

            # Read images (and figures)
            for image in reader.images.itertuples():
                namespace_handler["image"][image.tag].append(neptune.types.File.as_image(image.value))

            # Read text
            for text in reader.text.itertuples():
                namespace_handler["text"][text.tag].append(text.value)

            # Read hparams
            for hparam in reader.hparams.itertuples():
                namespace_handler["hparams"][hparam.tag].append(hparam.value)

            # user facing
            click.echo(f"{path} was exported with run_id: {hash_run_id}")



import os

import click


@click.command("tensorboard")
@click.option("--project", help="Neptune Project name")
@click.option("--api_token", help="Neptune API token")
@click.argument("log_dir", required=True)
def sync(project, api_token, log_dir):
    if not os.path.exists(log_dir):
        # user facing
        click.echo("ERROR: Provided `log_dir` path doesn't exist", err=True)
        return

    # We do not want to import anything if process was executed for autocompletion purposes.
    from neptune_tensorboard.sync import DataSync

    DataSync(project, api_token, log_dir).run()



import contextlib
import warnings
from functools import partial

import torch
from neptune.types import File
from neptune.utils import stringify_unsupported
from tensorboardX.writer import SummaryWriter

from neptune_tensorboard.integration.utils import register_pre_hook

IS_TORCHVIZ_AVAILABLE = True
try:
    import torchviz
except ImportError:
    IS_TORCHVIZ_AVAILABLE = False

__all__ = ["patch_tensorboardx", "NeptuneTensorboardXTracker"]

_integrated_with_tensorboardx = False


def patch_tensorboardx(run, base_namespace):
    global _integrated_with_tensorboardx

    if not _integrated_with_tensorboardx:
        NeptuneTensorboardXTracker(run, base_namespace)
        _integrated_with_tensorboardx = True


def track_scalar(
    summary_writer,
    tag,
    scalar_value,
    global_step=None,
    walltime=None,
    display_name="",
    summary_description="",
    run=None,
    base_namespace=None,
):
    run[base_namespace]["scalar"][tag].append(scalar_value)


def track_image(
    summary_writer, tag, img_tensor, global_step=None, walltime=None, dataformats="CHW", run=None, base_namespace=None
):
    if not isinstance(img_tensor, torch.Tensor):
        img_tensor = torch.tensor(img_tensor)
    # dataformats : CHW, HWC, HW, WH
    if dataformats == "CHW":
        # convert to HWC
        img_tensor = img_tensor.movedim(0, 2)
    elif dataformats == "WH":
        # convert to HW1
        img_tensor = img_tensor.movedim(0, 1).unsqueeze(2)
    elif dataformats == "HW":
        # convert to HW1
        img_tensor = img_tensor.unsqueeze(2)

    run[base_namespace]["image"][tag] = File.as_image(img_tensor)


def track_images(
    summary_writer, tag, img_tensor, global_step=None, walltime=None, dataformats="NCHW", run=None, base_namespace=None
):
    if not isinstance(img_tensor, torch.Tensor):
        img_tensor = torch.tensor(img_tensor)
    # dataformats : CHW, HWC, HW, WH
    if dataformats == "NCHW":
        # convert to HWC
        img_tensor = img_tensor.movedim(1, 3)
    elif dataformats == "NHWC":
        pass
    else:
        # user facing
        warnings.warn("neptune-tensorboard: Skipping logging images as  {dataformats} is not supported.")

    for idx in range(img_tensor.shape[0]):
        run[base_namespace]["images"][tag].append(File.as_image(img_tensor[idx]))


def track_figure(
    summary_writer, tag, figure, global_step=None, close=True, walltime=None, run=None, base_namespace=None
):
    run[base_namespace]["figure"][tag].append(figure)


def track_text(summary_writer, tag, text_string, global_step=None, walltime=None, run=None, base_namespace=None):
    run[base_namespace]["text"][tag].append(text_string)


def track_graph(
    summary_writer, model, input_to_model=None, verbose=False, use_strict_trace=True, run=None, base_namespace=None
):
    if not IS_TORCHVIZ_AVAILABLE:
        # user facing
        msg = "neptune-tensorboard: Skipping model visualization because no torchviz installation was found."
        warnings.warn(msg)
        return
    if input_to_model is None:
        # user facing
        msg = "neptune-tensorboard: Skipping model visualization because input_to_model was None."
        warnings.warn(msg)
        return
    output = model(input_to_model)
    graph = torchviz.make_dot(output, params=dict(model.named_parameters()))
    png_bytes = graph.pipe(format="png")
    run[base_namespace]["graph"].upload(File.from_content(png_bytes, extension="png"))


def track_hparam(
    summary_writer, hparam_dict, metric_dict, hparam_domain_discrete=None, run_name=None, run=None, base_namespace=None
):
    run[base_namespace]["hparams"] = stringify_unsupported(hparam_dict)
    run[base_namespace]["metrics"] = stringify_unsupported(metric_dict)


class NeptuneTensorboardXTracker(contextlib.AbstractContextManager):
    def __init__(self, run, base_namespace):
        self.org_add_scalar = SummaryWriter.add_scalar
        self.org_add_image = SummaryWriter.add_image
        self.org_add_images = SummaryWriter.add_images
        self.org_add_figure = SummaryWriter.add_figure
        self.org_add_text = SummaryWriter.add_text
        self.org_add_graph = SummaryWriter.add_graph
        self.org_add_hparams = SummaryWriter.add_hparams

        register_pre_hook_with_run = partial(register_pre_hook, run=run, base_namespace=base_namespace)

        SummaryWriter.add_scalar = register_pre_hook_with_run(
            original=SummaryWriter.add_scalar, neptune_hook=track_scalar
        )

        SummaryWriter.add_image = register_pre_hook_with_run(original=SummaryWriter.add_image, neptune_hook=track_image)

        SummaryWriter.add_images = register_pre_hook_with_run(
            original=SummaryWriter.add_images, neptune_hook=track_images
        )

        SummaryWriter.add_figure = register_pre_hook_with_run(
            original=SummaryWriter.add_figure, neptune_hook=track_figure
        )

        SummaryWriter.add_text = register_pre_hook_with_run(original=SummaryWriter.add_text, neptune_hook=track_text)

        SummaryWriter.add_graph = register_pre_hook_with_run(original=SummaryWriter.add_graph, neptune_hook=track_graph)

        SummaryWriter.add_hparams = register_pre_hook_with_run(
            original=SummaryWriter.add_hparams, neptune_hook=track_hparam
        )

    def __exit__(self, exc_type, exc_value, traceback):
        SummaryWriter.add_scalar = self.org_add_scalar
        SummaryWriter.add_image = self.org_add_image
        SummaryWriter.add_images = self.org_add_images
        SummaryWriter.add_figure = self.org_add_figure
        SummaryWriter.add_text = self.org_add_text
        SummaryWriter.add_graph = self.org_add_graph
        SummaryWriter.add_hparams = self.org_add_hparams



from functools import wraps


def register_pre_hook(original, neptune_hook, run, base_namespace):
    @wraps(original)
    def wrapper(*args, **kwargs):
        neptune_hook(*args, **kwargs, run=run, base_namespace=base_namespace)
        return original(*args, **kwargs)

    return wrapper
